@@ -5,15 +5,21 @@ Instances of these classes are called dataset objects.
 
 import logging
 
-from rasterio._base import (
-    get_dataset_driver, driver_can_create, driver_can_create_copy)
-from rasterio._io import (
-    DatasetReaderBase, DatasetWriterBase, BufferedDatasetWriterBase,
-    MemoryFileBase)
-from rasterio.windows import WindowMethodsMixin
-from rasterio.env import ensure_env, env_ctx_if_needed
-from rasterio.transform import TransformMethodsMixin
-from rasterio.path import UnparsedPath
+import rasterio._loading
+with rasterio._loading.add_gdal_dll_directories():
+    from rasterio._base import (
+        get_dataset_driver, driver_can_create, driver_can_create_copy)
+    from rasterio._io import (
+        DatasetReaderBase, DatasetWriterBase, BufferedDatasetWriterBase,
+        MemoryFileBase)
+    from rasterio.windows import WindowMethodsMixin
+    from rasterio.env import ensure_env
+    from rasterio.transform import TransformMethodsMixin
+    from rasterio._path import _UnparsedPath
+    try:
+        from rasterio._filepath import FilePathBase
+    except ImportError:
+        FilePathBase = object
 
 
 log = logging.getLogger(__name__)
@@ -122,10 +128,10 @@ class MemoryFile(MemoryFileBase):
         Other parameters are optional and have the same semantics as the
         parameters of `rasterio.open()`.
         """
-        mempath = UnparsedPath(self.name)
+        mempath = _UnparsedPath(self.name)
 
         if self.closed:
-            raise IOError("I/O operation on closed file.")
+            raise OSError("I/O operation on closed file.")
         if len(self) > 0:
             log.debug("VSI path: {}".format(mempath.path))
             return DatasetReader(mempath, driver=driver, sharing=sharing, **kwargs)
@@ -137,13 +143,91 @@ class MemoryFile(MemoryFileBase):
                           nodata=nodata, sharing=sharing, **kwargs)
 
     def __enter__(self):
-        self._env = env_ctx_if_needed()
-        self._env.__enter__()
         return self
 
-    def __exit__(self, *args, **kwargs):
+    def __exit__(self, *args):
         self.close()
-        self._env.__exit__()
+
+
+class _FilePath(FilePathBase):
+    """A BytesIO-like object, backed by an in-memory file.
+
+    This allows formatted files to be read and written without I/O.
+
+    A MemoryFile created with initial bytes becomes immutable. A
+    MemoryFile created without initial bytes may be written to using
+    either file-like or dataset interfaces.
+
+    Examples
+    --------
+
+    A GeoTIFF can be loaded in memory and accessed using the GeoTIFF
+    format driver
+
+    >>> with open('tests/data/RGB.byte.tif', 'rb') as f, FilePath(f) as vsi_file:
+    ...     with vsi_file.open() as src:
+    ...         pprint.pprint(src.profile)
+    ...
+    {'count': 3,
+     'crs': CRS({'init': 'epsg:32618'}),
+     'driver': 'GTiff',
+     'dtype': 'uint8',
+     'height': 718,
+     'interleave': 'pixel',
+     'nodata': 0.0,
+     'tiled': False,
+     'transform': Affine(300.0379266750948, 0.0, 101985.0,
+           0.0, -300.041782729805, 2826915.0),
+     'width': 791}
+
+    """
+
+    def __init__(self, filelike_obj, dirname=None, filename=None):
+        """Create a new wrapper around the provided file-like object.
+
+        Parameters
+        ----------
+        filelike_obj : file-like object
+            Open file-like object. Currently only reading is supported.
+        filename : str, optional
+            An optional filename. A unique one will otherwise be generated.
+
+        Returns
+        -------
+        PythonVSIFile
+        """
+        super().__init__(
+            filelike_obj, dirname=dirname, filename=filename
+        )
+
+    @ensure_env
+    def open(self, driver=None, sharing=False, **kwargs):
+        """Open the file and return a Rasterio dataset object.
+
+        The provided file-like object is assumed to be readable.
+        Writing is currently not supported.
+
+        Parameters are optional and have the same semantics as the
+        parameters of `rasterio.open()`.
+        """
+        mempath = _UnparsedPath(self.name)
+
+        if self.closed:
+            raise IOError("I/O operation on closed file.")
+        # Assume we were given a non-empty file-like object
+        log.debug("VSI path: {}".format(mempath.path))
+        return DatasetReader(mempath, driver=driver, sharing=sharing, **kwargs)
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, *args):
+        self.close()
+
+
+if FilePathBase is not object:
+    # only make this object available if the cython extension was compiled
+    FilePath = _FilePath
 
 
 class ZipMemoryFile(MemoryFile):
@@ -173,17 +257,17 @@ class ZipMemoryFile(MemoryFile):
         -------
         A Rasterio dataset object
         """
-        zippath = UnparsedPath('/vsizip{0}/{1}'.format(self.name, path.lstrip('/')))
+        zippath = _UnparsedPath('/vsizip{0}/{1}'.format(self.name, path.lstrip('/')))
 
         if self.closed:
-            raise IOError("I/O operation on closed file.")
+            raise OSError("I/O operation on closed file.")
         return DatasetReader(zippath, driver=driver, sharing=sharing, **kwargs)
 
 
 def get_writer_for_driver(driver):
     """Return the writer class appropriate for the specified driver."""
     if not driver:
-        raise ValueError("'driver' is required to write dataset.")
+        raise ValueError("'driver' is required to read/write dataset.")
     cls = None
     if driver_can_create(driver):
         cls = DatasetWriter
